using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace SmartphoneSnake
{
    public partial class SnakeForm : Form
    {
        public SnakeForm()
        {
            init();
            InitializeComponent();
        }

        #region Snake Position Variables

        /// <summary>
        /// Diameter of a dot in the snake. Set at a 
        /// DotWidthFactorth of the display width 
        /// - rounded down
        /// </summary>
        private int dotDiameter;

        /// <summary>
        /// Size of the dots relative to the screen width.
        /// A value of n means that n dots would fit in the screen
        /// width. 
        /// </summary>
        private const int DotWidthFactor = 20;

        /// <summary>
        /// X coordinate of top left hand corner of the
        /// snake head. Origin at left.
        /// </summary>
        private float snakeX;

        /// <summary>
        /// Y coordinate of top left hand corner of the 
        /// snake head. Origin at top.
        /// </summary>

        private float snakeY;
        /// <summary>
        /// Speed of snake in x direction.
        /// </summary>
        private float snakeXSpeed;

        /// <summary>
        /// Speed of snake in y direction.
        /// </summary>
        private float snakeYSpeed;

        /// <summary>
        /// The start speed of the snake.
        /// </summary>
        private float initialSnakeSpeed;

        /// <summary>
        /// The maximum speed of the snake.
        /// </summary>
        private float maximumSnakeSpeed;

        /// <summary>
        /// The change in speed that is applied to the snake.
        /// </summary>
        private float snakeSpeedStep;

        /// <summary>
        /// The speed of movement of the snake itself.
        /// </summary>
        private float snakeSpeed;

        /// <summary>
        /// Limit for the snake X value. Width of the screen
        /// minus the diameter of the snake dot.
        /// </summary>
        private float snakeXLimit;

        /// <summary>
        /// Limit for the snake Y value. Height of the screen 
        /// minus the diameter of the snake dot.
        /// </summary>
        private float snakeYLimit;

        #endregion 

        #region Snake Management


        #region Snake segment variables

        /// <summary>
        /// Keeps track of the position of a segment of the snake.
        /// </summary>
        class SnakeSegment
        {
            /// <summary>
            /// X position on the screen of the snake segment.
            /// </summary>
            public float X;

            /// <summary>
            /// Y position on the screen of the snake segment.
            /// </summary>
            public float Y;

            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float XSpeed;

            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float YSpeed;

            /// <summary>
            /// X offset to draw the snake segement (allows for segments
            /// of different sizes).
            /// </summary>
            public float XDrawOffset;

            /// <summary>
            /// Y offset to draw the snake segement (allows for segments
            /// of different sizes).
            /// </summary>
            public float YDrawOffset;

            /// <summary>
            /// Brush used to draw this snake segment.
            /// </summary>
            public Brush SnakeSegmentBrush;

            /// <summary>
            /// Rectangle defining the position of the snake on the screen.
            /// </summary>
            public Rectangle Rect;

            /// <summary>
            /// Draws the segment on the given graphics device
            /// </summary>
            /// <param name="g">device to be used for drawing</param>
            public void Draw ( Graphics g ) {
                Rect.X = (int)(X - XDrawOffset);
                Rect.Y = (int)(Y - YDrawOffset);
                g.FillRectangle(SnakeSegmentBrush, Rect);
            }

            /// <summary>
            /// Constructor. Do not set the position, since that
            /// will be set when the snake is placed on the screen.
            /// </summary>
            /// <param name="inBrush">brush to draw the snake segment with.</param>
            /// <param name="inDiameter">diameter of the snake segment.</param>
            public SnakeSegment(Brush inBrush, int inDiameter)
            {
                SnakeSegmentBrush = inBrush;
                XDrawOffset = inDiameter / 2.0f;
                YDrawOffset = inDiameter / 2.0f;
                Rect = new Rectangle(0, 0, inDiameter, inDiameter);
            }
        }

        /// <summary>
        /// The longest the snake is allowed to get.
        /// </summary>
        private static int maxSnakeLength = 40;

        /// <summary>
        /// The length of the snake at the start of the game.
        /// </summary>
        private int startSnakeLength = 30;

        /// <summary>
        /// The current length of the snake.
        /// </summary>
        private int currentSnakeLength ;

        /// <summary>
        /// The entire snake. Set up when the program is
        /// initialised. Use increasing amounts of the snake
        /// as the game progresses.
        /// </summary>
        private SnakeSegment[] wholeSnake ;

        #endregion 

        #region Snake management methods

        /// <summary>
        /// Initialise the snake ready for the game.
        /// </summary>
        private void initSnake()
        {
            wholeSnake = new SnakeSegment[maxSnakeLength];

            wholeSnake[0] = new SnakeSegment(snakeBrush, dotDiameter * 2);

            for (int segNo = 1; segNo < maxSnakeLength; segNo++)
            {
                wholeSnake[segNo] = new SnakeSegment(snakeBrush, dotDiameter); ;
            }
            currentSnakeLength = 0;
        }

        /// <summary>
        /// Erase the snake from the screen. Draw each block with a 
        /// background colour.
        /// </summary>
        /// <param name="g">Graphics context to draw on.</param>
        private void eraseSnake(Graphics g)
        {
            for (int segNo = 0; segNo < currentSnakeLength; segNo++)
            {
                SnakeSegment segment = wholeSnake[segNo];
                g.FillRectangle(backgroundBrush, segment.Rect);
            }
        }

        /// <summary>
        /// Draw the snake on the graphics surface given.
        /// </summary>
        /// <param name="g">Graphics context to draw on.</param>
        private void drawSnake(Graphics g)
        {
            for (int segNo = 0; segNo < currentSnakeLength; segNo++)
            {
                wholeSnake[segNo].Draw(g);
            }
        }


        /// <summary>
        /// Update the snake position. Move the head to the new position
        /// and move the body segments on to follow. Also increase the length 
        /// if the snake is growing.
        /// </summary>
        /// <param name="xHead">New head X position</param>
        /// <param name="yHead">New head Y position</param>
        private void updateSnakePosition ( 
            float xHead, float yHead, 
            float xSpeed, float ySpeed ) {

            SnakeSegment segmentInFront = wholeSnake[0]; ;

            if (currentSnakeLength < startSnakeLength)
            {
                currentSnakeLength++;
            }

            for (int i = currentSnakeLength - 1; i > 0 ; i--)
            {
                SnakeSegment current = wholeSnake[i];
                segmentInFront = wholeSnake[i - 1];
                current.X = segmentInFront.X;
                current.Y = segmentInFront.Y;
                current.XSpeed = segmentInFront.XSpeed;
                current.YSpeed = segmentInFront.YSpeed;
            }

            // at the end of the loop segmentInFront points at 
            // the head segment

            segmentInFront.X = xHead;
            segmentInFront.Y = yHead;
            segmentInFront.XSpeed = xSpeed;
            segmentInFront.YSpeed = ySpeed;
        }

        /// <summary>
        /// Checks to see if the snake is overwriting itself.
        /// </summary>
        /// <returns>true if the snake crashes into itself.</returns>
        private bool snakeCrash()
        {
            SnakeSegment head = wholeSnake[0];

            for (int segNo = 1; segNo < currentSnakeLength; segNo++)
            {
                SnakeSegment segment = wholeSnake[segNo];
                if (head.Rect.IntersectsWith(segment.Rect)) {
                    ///only have a crash if the head segment is 
                    ///moving towards this segment
                    float xDir = (head.X - segment.X) * head.XSpeed;
                    float yDir = (head.Y - segment.Y) * head.YSpeed;
                    if (xDir < 0 || yDir < 0 )
                    {
                         return true;
                    }
                }
            }
            return false;
        }

        #endregion

        #endregion

        #region Snake Drawing variables

        /// <summary>
        /// Brush used to draw the snake body. Kept as a member
        /// to avoid repeated recreation. Created in init.
        /// </summary>
        private Brush snakeBrush;

        /// <summary>
        /// Brush used to draw the background. Kept as a member
        /// to avoid repeated recreation. Created in init.
        /// </summary>
        private Brush backgroundBrush;

        /// <summary>
        /// Graphics device for this form. Used to draw all
        /// screen items. Assigned in init.
        /// </summary>
        private Graphics gameGraphics;

        #endregion

        #region Game State variables

        /// <summary>
        /// When true the game is being played.
        /// </summary>
        private bool gameActive = false;

        /// <summary>
        /// When true the game is in pause state.
        /// </summary>
        private bool gamePaused = false;

        #endregion

        /// <summary>
        /// Sets up the game at the start.
        /// </summary>
        private void init()
        {
            snakeBrush = new SolidBrush(Color.Red);
            backgroundBrush = new SolidBrush(Color.White);

            gameGraphics = this.CreateGraphics();

            dotDiameter = this.ClientRectangle.Width / DotWidthFactor;

            snakeX = (this.ClientRectangle.Width - dotDiameter) / 2;
            snakeY = (this.ClientRectangle.Height - dotDiameter) / 2;

            initialSnakeSpeed = (float)dotDiameter / 1.5f;
            maximumSnakeSpeed = (float)dotDiameter / 2.0f;
            snakeSpeedStep = (maximumSnakeSpeed - initialSnakeSpeed) / 5;

            snakeSpeed = initialSnakeSpeed;

            snakeXSpeed = snakeSpeed;
            snakeYSpeed = 0;

            snakeXLimit = this.ClientRectangle.Width - dotDiameter;
            snakeYLimit = this.ClientRectangle.Height - dotDiameter;

            initSnake();

        }

        /// <summary>
        /// Update the snake position. End the game if
        /// it hits anything.
        /// </summary>
        private void updateSnake()
        {
            snakeX += snakeXSpeed;
            snakeY += snakeYSpeed;
            if ((snakeX < 0.0f) || (snakeX > snakeXLimit) ||
                 (snakeY < 0.0f) || (snakeY > snakeYLimit)  ||
                 snakeCrash() )
            {
                MessageBox.Show("Game Over");
                eraseGameScreen();
                gameActive = false;
            }

            updateSnakePosition(snakeX, snakeY, snakeXSpeed, snakeYSpeed);

        }

        /// <summary>
        /// draw the game screen using the graphics for this form.
        /// </summary>
        private void drawGameScreen()
        {
            drawSnake(gameGraphics);
        }

        /// <summary>
        /// Erases the game screen using the graphics for this form.
        /// </summary>
        private void eraseGameScreen()
        {
            eraseSnake(gameGraphics);
        }

        #region Game Play

        /// <summary>
        /// Plays the game. Sets gameActive to true and
        /// updates the game until it becomes false.
        /// </summary>
        private void playGame()
        {
            init();
            gameActive = true;
            while (gameActive)
            {
                if (!gamePaused)
                {
                    eraseGameScreen();
                    updateSnake();
                    drawGameScreen();
                }
                System.Threading.Thread.Sleep(100);
                Application.DoEvents();
            }
        }

        /// <summary>
        /// Refers to a method with the signature required for 
        /// Form.Invoke calls.
        /// </summary>
        private delegate void gameMethod();

        /// <summary>
        /// Refers to the method which will play the game. Set
        /// to playGame when the game starts.
        /// </summary>
        gameMethod playDelegate;

        /// <summary>
        /// Starts a new game.
        /// </summary>
        private void startGame()
        {
            playDelegate = new gameMethod(playGame);
            Invoke(playDelegate);
        }

        /// <summary>
        /// Pause the game.
        /// </summary>
        private void pauseGame()
        {
            gamePaused = true;
        }

        /// <summary>
        /// Resume the game
        /// </summary>
        private void resumeGame()
        {
            gamePaused = false;
        }

        #endregion 

        #region Event Handlers

        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);
            pauseGame();
        }

        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);
            resumeGame();
        }


        protected override void OnPaint(PaintEventArgs e)
        {
            drawGameScreen();
        }

        private void playMenuItem_Click(object sender, EventArgs e)
        {
            startGame();
        }

        private void SnakeForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((e.KeyCode == System.Windows.Forms.Keys.F1))
            {
                // Soft Key 1
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F2))
            {
                // Soft Key 2
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Up))
            {
                // Up
                snakeXSpeed = 0.0f;
                snakeYSpeed = -snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Down))
            {
                // Down
                snakeXSpeed = 0.0f;
                snakeYSpeed = snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Left))
            {
                // Left
                snakeXSpeed = -snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Right))
            {
                // Right
                snakeXSpeed = snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Enter))
            {
                // Enter
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D1))
            {
                // 1
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D2))
            {
                // 2
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D3))
            {
                // 3
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D4))
            {
                // 4
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D5))
            {
                // 5
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D6))
            {
                // 6
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D7))
            {
                // 7
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D8))
            {
                // 8
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D9))
            {
                // 9
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F8))
            {
                // *
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D0))
            {
                // 0
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F9))
            {
                // #
            }
        }

        #endregion 
    }
}