using System;
using System.Drawing;

namespace PlayWithFriends
{
	/// <summary>
	/// This is the root class of all the sprites defined 
	/// in the game.
	/// </summary>
	public abstract class BaseSprite
	{
		/// <summary>
		/// Location of the sprite in the screen area. The position of the rectangle
		/// is managed relative to playfield scrolling.
		/// </summary>
		public Rectangle SpriteRectangle;

		/// <summary>
		/// Limits of the screen drawing area.
		/// </summary>
		public Rectangle BoundingRectangle;

		/// <summary>
		/// X position on the drawing area.
		/// </summary>
		public int X;

		/// <summary>
		/// Y position on the drawing area.
		/// </summary>
		public int Y;

		/// <summary>
		/// Random number for the use of sprites. Seeded so that the game is always
		/// given the same sequence each tiem.
		/// </summary>
		protected static Random rand = new Random(1);

		/// <summary>
		/// Draws the current sprite image.
		/// </summary>
		/// <param name="g">graphics context for the draw operation</param>
		public abstract void Draw ( Graphics g) ;

		/// <summary>
		/// Update the sprite position and state. 
		/// </summary>
		public abstract void Update () ;

		/// <summary>
		/// Reset to initial position and state.
		/// </summary>
		public virtual void Reset () 
		{
			// do nothing at the moment
		}

		/// <summary>
		/// Base constructor.
		/// </summary>
		/// <param name="inSpriteRectangle">Bounding rectangle for the sprite.</param>
		/// <param name="inBoundingRectangle">Bounding rectangle for the display area.</param>
		public BaseSprite( 
			Rectangle inSpriteRectangle, 
			Rectangle inBoundingRectangle )
		{
			SpriteRectangle = inSpriteRectangle;
			BoundingRectangle = inBoundingRectangle;
		}
	}

	#region Background sprites

	/// <summary>
	/// Simple background star. Just moves down the screen.
	/// </summary>
	public class BackgroundStarSprite : BaseSprite
	{
		public const int backgroundStarWidth = 1;
		public const int backgroundStarHeight = 1;
		public const int backgroundStarSpeed = 2;

		private static Brush backgroundStarBrush = new SolidBrush(Color.White);

		public override void Draw(Graphics g)
		{
			g.FillRectangle( backgroundStarBrush, SpriteRectangle ); 
		}

		public override void Update()
		{
			Y += backgroundStarSpeed ;
			if (Y > BoundingRectangle.Height ) 
			{
				X = BaseSprite.rand.Next(BoundingRectangle.Width - backgroundStarWidth);
				Y = 0 ;
			}
		}

		public override void Reset()
		{
			base.Reset();
			X = BaseSprite.rand.Next(BoundingRectangle.Width - backgroundStarWidth);
			Y = BaseSprite.rand.Next(BoundingRectangle.Height - backgroundStarHeight);
		}

		public BackgroundStarSprite ( 
			Rectangle inBoundingRectangle ) : 
			base ( 
			new Rectangle(
			BaseSprite.rand.Next(inBoundingRectangle.Width - backgroundStarWidth),
			BaseSprite.rand.Next(inBoundingRectangle.Height - backgroundStarHeight),
			backgroundStarWidth,
			backgroundStarHeight ),
			inBoundingRectangle )
		{
			base.Reset();
			Reset();
		}
	}

	/// <summary>
	/// Draws a colored sprite which changes color as it moves down the screen. Moves
	/// faster than the background sprites to give parallax scrolling.
	/// </summary>
	public class ColoredBackgroundStarSprite : BaseSprite
	{
		public const int coloredBackgroundStarWidth = 3;
		public const int coloredBackgroundStarHeight = 3;
		public const int coloredBackgroundStarSpeed = 3;

		private int colorCount;
		private int flickerCount;
		private int flickerLimit;

		private static Brush [] buildBrushes (int seed) 
		{
			int size = 5 + rand.Next(10);
			int red;
			int green;
			int blue;
			Brush [] result = new Brush [size];
			for ( int i=0 ; i < size ; i++ ) 
			{
				red = 127 + (64*rand.Next(3));
				blue = 127 + (64*rand.Next(3));
				green = 63 + (64*rand.Next(4));

				result[i] = new SolidBrush(Color.FromArgb( red, green, blue ) );
			}
			return result;
		}

		private static Brush [] spriteBrushes = buildBrushes ( 1 );

		private static Brush brush = new SolidBrush(Color.White);

		public override void Draw(Graphics g)
		{
			g.FillRectangle( spriteBrushes[colorCount], SpriteRectangle ); 
			if ( ++flickerCount > flickerLimit ) 
			{
				flickerCount = 0;
				flickerLimit = 10 + rand.Next(100);
				colorCount++;
				if (colorCount == spriteBrushes.Length) 
				{
					colorCount = 0 ;
				}
			}
		}

		public override void Update()
		{
			Y += coloredBackgroundStarSpeed;
			if ( Y > BoundingRectangle.Height ) 
			{
				X = BaseSprite.rand.Next(BoundingRectangle.Width - coloredBackgroundStarWidth );
				Y = 0 ;
			}
		}

		public override void Reset()
		{
			base.Reset();
			X = BaseSprite.rand.Next(BoundingRectangle.Width - coloredBackgroundStarWidth ) ;
			Y = BaseSprite.rand.Next(BoundingRectangle.Height - coloredBackgroundStarHeight ) ;
		}

		public ColoredBackgroundStarSprite ( 
			Rectangle inBoundingRectangle ) : 
			base ( 
			new Rectangle(
			BaseSprite.rand.Next(inBoundingRectangle.Width - coloredBackgroundStarWidth),
			BaseSprite.rand.Next(inBoundingRectangle.Height - coloredBackgroundStarHeight),
			coloredBackgroundStarWidth,
			coloredBackgroundStarHeight ),
			inBoundingRectangle )
		{
			Reset();
		}
	}

	#endregion

	#region Image sprites

	/// <summary>
	/// Animated bitmap for use in sprites which change their display. Converts the
	/// bitmap as supplied into a grid of images which can be selected to animate 
	/// the display.
	/// </summary>
	public class AnimatedBitmap 
	{
		/// <summary>
		/// Bitmap holding all the sprite images. Uses the top row and left hand column 
		/// as height indicators for the cells in the bitmap
		/// </summary>
		private Bitmap spriteBitmap ;

		/// <summary>
		/// Width of a single frame in the animation area. Computed using the rulers in 
		/// the bitmap.
		/// </summary>
		public int FrameWidth;

		/// <summary>
		/// Height of a single frame in the animation area. Computed using the rulers
		/// in the bitmap.
		/// </summary>
		public int FrameHeight;

		/// <summary>
		/// Number of frames in the animation.
		/// </summary>
		public int FrameCount ;

		/// <summary>
		/// Number of animation rows.
		/// </summary>
		public int RowCount ;

		/// <summary>
		/// Precomputed x offests for the draw operations. 
		/// </summary>
		private int [] xOffsets ;
		/// <summary>
		/// Precomputed y offsets for the draw operations.
		/// </summary>
		private int [] yOffsets ;

		/// <summary>
		/// Property for the bitmap. Makes sure that the width and height
		/// and offsets are computed when the bitmap is changed.
		/// </summary>
		public Bitmap SpriteBitmap 
		{
			get 
			{
				return spriteBitmap;
			}
			set 
			{
				spriteBitmap = value;
				setWidthAndHeight ();
			}
		}

		/// <summary>
		/// Internal frame number being displayed
		/// </summary>
		private int frameNo ;

		/// <summary>
		/// Property for frame number. Note that it wraps round so that continuous
		/// increasing or decreasing sequences correctly.
		/// </summary>
		public int FrameNo 
		{
			get 
			{
				return frameNo;
			}
			set 
			{
				if ( value < 0 ) 
				{
					frameNo = FrameCount - 1;
					return;
				}
				if ( value >= FrameCount ) 
				{
					frameNo = 0;
					return;
				}
				frameNo = value;
			}
		}

		/// <summary>
		/// Number for the row being displayed.
		/// </summary>
		private int rowNo ;

		/// <summary>
		/// Property for the row number. Note that this does not wrap around as the
		/// frame does, as it would not be useful in this context.
		/// </summary>
		public int RowNo 
		{
			get 
			{
				return rowNo;
			}
			set 
			{
				if ( value < 0 ) 
				{
					rowNo = 0;
					return;
				}
				if ( value >= RowCount ) 
				{
					rowNo = RowCount - 1;
					return;
				}
				rowNo = value;
			}
		}

		/// <summary>
		/// Computes the width and height and offests for the given image.
		/// </summary>
		private void setWidthAndHeight () 
		{
			// color on the top left hand corner gives the color for the ruler bars

			Color searchColor = spriteBitmap.GetPixel(0,0);

			FrameWidth = 0;
			while ( 
				( spriteBitmap.GetPixel ( FrameWidth, 0 ) == searchColor ) &&
				( FrameWidth < spriteBitmap.Width ) )
			{
				FrameWidth++;
			}

			FrameHeight = 0;
			while ( 
				( spriteBitmap.GetPixel ( 0, FrameHeight ) == searchColor ) &&
				( FrameHeight < spriteBitmap.Height ) )
			{
				FrameHeight++;
			}

			FrameCount = ( spriteBitmap.Width - 1 ) / FrameWidth ;
			RowCount = ( spriteBitmap.Height - 1 ) / FrameHeight ;

			xOffsets = new int [ FrameCount ] ;
			yOffsets = new int [ RowCount ] ;

			xOffsets[0] = 1;
			yOffsets[0] = 1;

			for ( int i=1; i < FrameCount; i++ ) 
			{
				xOffsets[i] = xOffsets[i-1] + FrameWidth;
			}

			for ( int i=1; i < RowCount; i++ ) 
			{
				yOffsets[i] = yOffsets[i-1] + FrameHeight;
			}
		}

		/// <summary>
		/// Draw teh required frame in the appropriate place.
		/// </summary>
		/// <param name="g">graphics to draw with</param>
		/// <param name="spriteRectangle">destination rectangle</param>
		public void Draw(Graphics g, Rectangle spriteRectangle)
		{
			g.DrawImage(
				spriteBitmap,
				spriteRectangle,
				xOffsets[frameNo], 
				yOffsets[rowNo],
				FrameWidth, FrameHeight,
				GraphicsUnit.Pixel,
				Settings.TransparentAttributes );
		}

		/// <summary>
		/// Create an instace from the bitmap supplied.
		/// </summary>
		/// <param name="inSpriteBitmap">the bitmap, with the size rulers in it</param>
		public AnimatedBitmap( Bitmap inSpriteBitmap ) 
		{
			SpriteBitmap = inSpriteBitmap;
		}

	}

	/// <summary>
	/// The states that the ship can occupy.
	/// </summary>
	public enum ShipSpriteStates 
	{
		visible, 
		hit,
		exploding,
		invisible
	}

	/// <summary>
	/// The sprite for the ship. Keeps track of ship health.
	/// Provides a method which allows it to receive damage and
	/// explode.
	/// </summary>
	public class ShipSprite : BaseSprite 
	{
		/// <summary>
		/// Animated bitmap for normal ship display.
		/// </summary>
		public AnimatedBitmap SpriteAnimatedBitmap ;

		/// <summary>
		/// Animation played when the ship explodes.
		/// </summary>
		public  AnimatedBitmap ExplosionAnimatedBimap ;

		#region State management

		/// <summary>
		/// State of the ship
		/// </summary>
		private ShipSpriteStates state ;

		public ShipSpriteStates State 
		{
			get 
			{
				return state;
			}
			set 
			{
				switch ( value ) 
				{
					case ShipSpriteStates.exploding:
						break;
					case ShipSpriteStates.hit:
						break;
					case ShipSpriteStates.invisible:
						break;
					case ShipSpriteStates.visible:
						break;
				}
				state = value;
			}
		}

		#endregion

		#region Position management

		/// <summary>
		/// X update value. To move the ship the controller must put values
		/// in these members. They are applied to the position of the
		/// ship next time it is updated.
		/// </summary>
		public int XUpdateValue = 0;

		/// <summary>
		/// Y update value. Changed to move the ship up and down.
		/// </summary>
		public int YUpdateValue = 0;

		#endregion

		#region Overridden ShipSprite methods
		
		/// <summary>
		/// Updates the position of the ship by applying change values which have
		/// been supplied. Performs range checking to keep the sprite on the
		/// screen.
		/// </summary>
		public override void Update()
		{
			X += XUpdateValue;
			if ( X < 0 ) 
			{
				X = 0;
			}
			if ( (X + SpriteAnimatedBitmap.FrameWidth) > BoundingRectangle.Width ) 
			{
				X = ( BoundingRectangle.Width - SpriteAnimatedBitmap.FrameWidth ); 
			}
																								
			Y += YUpdateValue;
			if ( Y < 0 ) 
			{
				Y = 0 ;
			}
			if ( (Y + SpriteAnimatedBitmap.FrameHeight) > BoundingRectangle.Height ) 
			{
				Y = ( BoundingRectangle.Height - SpriteAnimatedBitmap.FrameHeight ); 
			}

			XUpdateValue = 0;
			YUpdateValue = 0;
		}

		/// <summary>
		/// This is used to draw the protective halo around the ship when it has been
		/// hit.
		/// </summary>
		private Brush hitBrush = new SolidBrush ( Color.Yellow );

		/// <summary>
		/// Draws the ship according to the current state. May also perform a state
		/// change if the end of an animation sequence has been reached.
		/// </summary>
		/// <param name="g">graphics to use for the drawing.</param>
		public override void Draw(Graphics g)
		{

			switch (State) 
			{
				case ShipSpriteStates.visible :
					SpriteAnimatedBitmap.Draw(g, SpriteRectangle);
					SpriteAnimatedBitmap.FrameNo++;
					break;
				case ShipSpriteStates.hit :
					g.FillEllipse( 
						hitBrush, 
						SpriteRectangle.X+2, 
						SpriteRectangle.Y+2, 
						SpriteRectangle.Width-4, 
						SpriteRectangle.Height-4 ) ;
					SpriteAnimatedBitmap.Draw(g, SpriteRectangle);
					SpriteAnimatedBitmap.FrameNo++;

					invulnerableCount++ ;
					if ( invulnerableCount >= invulnerableLimit ) 
					{
						invulnerableCount = 0;
						state = ShipSpriteStates.visible ;
					}
					break;
				case ShipSpriteStates.exploding :
					ExplosionAnimatedBimap.Draw(g, SpriteRectangle);
					ExplosionAnimatedBimap.FrameNo++;

					if (ExplosionAnimatedBimap.FrameNo == 0) 
					{
						State = ShipSpriteStates.invisible;
					}
					break ;
				case ShipSpriteStates.invisible :
					break ;
			}
		}

		/// <summary>
		/// Put the ship back at the bottom of the screen and reset all counters.
		/// </summary>
		public override void Reset()
		{
			base.Reset();
			hitCount = 0 ;
			invulnerableCount = 0 ;
			SpriteAnimatedBitmap.RowNo = 0 ;
			state = ShipSpriteStates.visible ;
			X = (BoundingRectangle.Width - SpriteRectangle.Width) / 2 ;
			Y = (BoundingRectangle.Height - SpriteRectangle.Height) ;
		}

		#endregion

		#region Code dealing with damage

		/// <summary>
		/// Timer for the period of invulnerability after a hit.
		/// </summary>
		private int invulnerableCount = 0;

		/// <summary>
		/// Limit for the period of invulnerability after a hit.
		/// </summary>
		private int invulnerableLimit = 30;

		/// <summary>
		/// Sound made when we are hit but not destroyed
		/// </summary>
		private Sound hitSound;

		/// <summary>
		/// Sound made when we are destroyed
		/// </summary>
		private Sound explodeSound;

		/// <summary>
		/// Number of times the ship has been hit.
		/// </summary>
		private int hitCount;

		/// <summary>
		/// Number of times the ship can be hit before it explodes.
		/// </summary>
		public int HitLimit = 3;

		/// <summary>
		/// Method called to deliver a hit to the ship. 
		/// </summary>
		/// <param name="power">the amount of the damage</param>
		/// <returns>true if the hit has killed the ship</returns>
		public bool HitHasKilled ( int power ) 
		{
			hitCount += power ;

			SpriteAnimatedBitmap.RowNo = hitCount;

			if ( hitCount >= HitLimit ) 
			{
				hitCount = 0;
				State = ShipSpriteStates.exploding;
				explodeSound.Play();
				return true;
			}
			else 
			{
				State = ShipSpriteStates.hit;
				hitSound.Play();
			}
			return false;
		}

		#endregion

		#region Constructor

		public ShipSprite ( 
			Bitmap inSpriteBitmap, 
			Bitmap inExplodingBitmap,
			Sound inHitSound,
			Sound inExplodeSound,
			Rectangle inBoundingRectangle ) :
			base ( new Rectangle (0,0,10,10), inBoundingRectangle ) 
		{
			SpriteAnimatedBitmap = new AnimatedBitmap ( inSpriteBitmap );
			ExplosionAnimatedBimap = new AnimatedBitmap ( inExplodingBitmap );
			hitSound = inHitSound;
			explodeSound = inExplodeSound;
			SpriteRectangle.Width = SpriteAnimatedBitmap.FrameWidth;
			SpriteRectangle.Height = SpriteAnimatedBitmap.FrameHeight;
			Reset();
		}

		#endregion

	}

	#endregion

	#region Projectile sprites

	/// <summary>
	/// The states that a missile sprite can occupy.
	/// </summary>
	public enum MissileSpriteStates
	{
		loaded,
		flying
	}

	/// <summary>
	/// Displays a missile. A missile is a simple rectangle of a particular color
	/// and size.
	/// </summary>
	public class MissileSprite : BaseSprite 
	{
		/// <summary>
		/// Used to draw the missile on the screen. Public so that 
		/// it can be changed to reflect missile upgrades.
		/// </summary>
		public Brush missileBrush ;

		/// <summary>
		/// Speed the missile moves each step. Public so that it
		/// can be updated to reflect missile upgrades.
		/// </summary>
		public int MissileSpeed = 5 ;

		/// <summary>
		/// Damage the missile inflicts on the enemy. Initially only
		/// one hit. May be increased with upgrades.
		/// </summary>
		public int MissilePower = 1 ;

		/// <summary>
		/// Sound the missile makes when fired
		/// </summary>
		private Sound fireSound;
		
		/// <summary>
		/// Internal state value.
		/// </summary>
		private MissileSpriteStates state ;

		/// <summary>
		/// External state property.
		/// </summary>
		public MissileSpriteStates State
		{
			get
			{
				return state;
			}
			set
			{
				state = value;
			}
		}

		/// <summary>
		/// Only draws the missile if it is in the air.
		/// </summary>
		/// <param name="g">graphics to be used for drawing.</param>
		public override void Draw(Graphics g)
		{
			switch ( state ) 
			{
				case MissileSpriteStates.flying:
					g.FillRectangle( missileBrush, SpriteRectangle ); 
					break;
				case MissileSpriteStates.loaded:
					break;
			}
		}

		/// <summary>
		/// Updates the missile movement. At the moment it just moves up the screen.
		/// If the missile reaches the end of the screen it is reloaded, so that it 
		/// can be fired again. Note that  the Y origin is the top of the screen.
		/// </summary>
		public override void Update()
		{
			switch ( state ) 
			{
				case MissileSpriteStates.flying:
					Y -= MissileSpeed;
					if ( Y < 0 ) 
					{
						state = MissileSpriteStates.loaded;
					}
					break;
				case MissileSpriteStates.loaded:
					break;
			}
		}

		/// <summary>
		/// reloads the missile
		/// </summary>
		public override void Reset()
		{
			base.Reset();
			state = MissileSpriteStates.loaded;
		}

		/// <summary>
		/// Fires the missile. Puts it on the screen and sets it moving. Will not
		/// do anything if the missile is already in the air.
		/// </summary>
		/// <param name="x">initial x position.</param>
		/// <param name="y">initial y position.</param>
		public void Fire ( int x, int y) 
		{
			if ( state != MissileSpriteStates.loaded ) 
			{
				return;
			}

			X = x;
			Y = y;
			state = MissileSpriteStates.flying;
			fireSound.Play();

		}

		public MissileSprite ( 
			Brush inMissileBrush, 
			Sound inFireSound,
			Rectangle inMissileRect, 
			Rectangle inBoundingRectangle ) : 
			base ( 
			inMissileRect, 
			inBoundingRectangle )
		{
			missileBrush = inMissileBrush;
			fireSound = inFireSound;
			Reset();
		}
	}

	#endregion

	#region Scripted sprites

	public abstract class ScriptedSprite : BaseSprite 
	{
		public static ShipSprite ship ;
		public static MissileSprite missile ;

		public bool IsAlive ;

		// general purpose tag for any information the sprite needs
		public object Tag;

		// floating point position values used in stepping 
		public float Xfloat; 
		public float Yfloat;

		// general purpose counter for use by scripts
		public int Counter;

		/// <summary>
		/// Current position in the script the sprite is executing
		/// </summary>
		public SpriteScriptStep CurrentSpriteSciptStep ;

		/// <summary>
		/// The start position in the script for this sprite
		/// </summary>
		public SpriteScriptStep ResetSpriteScriptStep;

		/// <summary>
		/// The start X position on the screen for this sprite
		/// </summary>
		public int ResetSpritePositionX;
		/// <summary>
		/// The start Y position on the screen for this sprite
		/// </summary>
		public int ResetSpritePositionY;

		/// <summary>
		/// Current attack wave the sprite is part of
		/// </summary>
		public AttackWave CurrentAttackWave ;

		public override void Update()
		{
			if (  CurrentSpriteSciptStep != null ) 
			{
				CurrentSpriteSciptStep.Command ( this );
			}
		}

		abstract public bool Visible 
		{
			get;
			set;
		}

		public override void Reset()
		{
			CurrentSpriteSciptStep = ResetSpriteScriptStep;
			X = ResetSpritePositionX;
			Y = ResetSpritePositionY;
			IsAlive = true;
			Counter = 0 ;
		}


		public ScriptedSprite( 
			Rectangle inSpriteRectangle, 
			Rectangle inBoundingRectangle ) : base ( inSpriteRectangle, inBoundingRectangle )
		{
		}
	}

	#endregion

	#region Alien sprites

	public enum AlienSpriteStates 
	{
		exploding,
		hit,
		visible,
		invisible
	}

    public abstract class AlienSprite : ScriptedSprite
    {
        /// <summary>
        /// Animation played when the alien explodes.
        /// </summary>
        public AnimatedBitmap ExplosionAnimatedBitmap;

        /// <summary>
        /// The number of hits sustained by the alien.
        /// </summary>
        protected int hitCount;

        /// <summary>
        /// Number of times the alien can be hit before it explodes.
        /// </summary>
        public int HitLimit = 2;

        /// <summary>
        /// Sound made when the alien is hit.
        /// </summary>
        protected Sound hitSound;

        /// <summary>
        /// Sound made when the alien explodes.
        /// </summary>
        protected Sound explodeSound;

        /// <summary>
        /// Manages the incoming hit.
        /// </summary>
        /// <param name="power">power of the hit</param>
        /// <returns>true if the hit has killed the alien</returns>
        public virtual bool HitHasKilled(int power)
        {
            hitCount += power;

            if (hitCount >= HitLimit)
            {
                hitCount = 0;
                State = AlienSpriteStates.exploding;
                return true;
            }
            else
            {
                State = AlienSpriteStates.hit;
                hitSound.Play();
            }
            return false;
        }

        // state management

        protected AlienSpriteStates state;

        public AlienSpriteStates State
        {
            get
            {
                return state;
            }
            set
            {
                switch (value)
                {
                    case AlienSpriteStates.exploding:
                        explodeSound.Play();
                        break;
                    case AlienSpriteStates.invisible:
                        break;
                    case AlienSpriteStates.visible:
                        break;
                }
                state = value;
            }
        }

        public override bool Visible
        {
            get
            {
                return (state == AlienSpriteStates.visible);
            }
            set
            {
                if (value)
                {
                    state = AlienSpriteStates.visible;
                }
                else
                {
                    state = AlienSpriteStates.invisible;
                }
            }
        }

        protected Brush hitBrush = new SolidBrush(Color.Red);

        public AlienSprite(
            Bitmap inExplodeBitmap,
            Sound inHitSound,
            Sound inExplodeSound,
            Rectangle inBoundingRectangle)
            :
            base(new Rectangle(0, 0, 10, 10), inBoundingRectangle)
        {
            ExplosionAnimatedBitmap = new AnimatedBitmap(inExplodeBitmap);
            hitSound = inHitSound;
            explodeSound = inExplodeSound;
        }

    }

	/// <summary>
	/// The alien sprite. Can be controlled by a script and contains animated bitmaps
	/// for display and explosion.
	/// </summary>
    public class AnimatedAlienSprite : AlienSprite 
	{
		/// <summary>
		/// Animated bitmap for normal alien display.
		/// </summary>
		public AnimatedBitmap SpriteAnimatedBitmap ;

		public override void Draw(Graphics g)
		{
			switch (State) 
			{
				case AlienSpriteStates.visible :
					SpriteAnimatedBitmap.Draw(g, SpriteRectangle);
					SpriteAnimatedBitmap.FrameNo++;
					break;
				case AlienSpriteStates.hit :
					g.FillEllipse( 
						hitBrush, 
						SpriteRectangle.X+2, 
						SpriteRectangle.Y+2, 
						SpriteRectangle.Width-4, 
						SpriteRectangle.Height-4 ) ;
					SpriteAnimatedBitmap.Draw(g, SpriteRectangle);
					state = AlienSpriteStates.visible ;
					break;
				case AlienSpriteStates.exploding :
					ExplosionAnimatedBitmap.Draw(g, SpriteRectangle);
					ExplosionAnimatedBitmap.FrameNo++;

					if (ExplosionAnimatedBitmap.FrameNo == 0) 
					{
						State = AlienSpriteStates.invisible;
						IsAlive = false;
					}
					break ;
				case AlienSpriteStates.invisible :
					break ;
			}
		}

        public override bool HitHasKilled(int power)
        {
            SpriteAnimatedBitmap.RowNo = hitCount;
            return base.HitHasKilled(power);
        }

		public override void Reset()
		{
			base.Reset();
			hitCount = 0 ;
			SpriteAnimatedBitmap.RowNo = 0;
			state = AlienSpriteStates.invisible;
		}


		public AnimatedAlienSprite ( 
			Bitmap inSpriteBitmap, 
			Bitmap inExplodeBitmap,
			Sound inHitSound,
			Sound inExplodeSound,
			Rectangle inBoundingRectangle ) : base ( inExplodeBitmap, inHitSound,inExplodeSound,inBoundingRectangle)
		{
			SpriteAnimatedBitmap = new AnimatedBitmap ( inSpriteBitmap );
			SpriteRectangle.Width = SpriteAnimatedBitmap.FrameWidth;
			SpriteRectangle.Height = SpriteAnimatedBitmap.FrameHeight;
			Reset();
		}
	}

    /// <summary>
    /// The alien sprite. Can be controlled by a script and contains animated bitmaps
    /// for display and explosion.
    /// </summary>
    public class ImageAlienSprite : AlienSprite
    {
        /// <summary>
        /// Bitmap for normal alien display.
        /// </summary>
        private Bitmap spriteBitmapValue;

        public Bitmap SpriteBitmap
        {
            get { 
                return spriteBitmapValue; 
            }
            set { 
                spriteBitmapValue = value;

                SpriteRectangle.Width = spriteBitmapValue.Width;
                SpriteRectangle.Height = spriteBitmapValue.Height;
                transparentAttributes.ClearColorKey();
                transparentAttributes.SetColorKey(
                    spriteBitmapValue.GetPixel(0, 0),
                    spriteBitmapValue.GetPixel(0, 0));
            }
        }

        /// <summary>
        /// Transparent draw attributes for the image
        /// </summary>
        private System.Drawing.Imaging.ImageAttributes 
            transparentAttributes = new System.Drawing.Imaging.ImageAttributes();

        public override void Draw(Graphics g)
        {
            switch (State)
            {
                case AlienSpriteStates.visible:
                    g.DrawImage(
                        SpriteBitmap,
                        SpriteRectangle,
                        0,
                        0,
                        SpriteBitmap.Width, SpriteBitmap.Height,
                        GraphicsUnit.Pixel,
                        transparentAttributes);

                    break;
                case AlienSpriteStates.hit:
                    g.FillEllipse(
                        hitBrush,
                        SpriteRectangle.X + 2,
                        SpriteRectangle.Y + 2,
                        SpriteRectangle.Width - 4,
                        SpriteRectangle.Height - 4);
                    g.DrawImage(
                        SpriteBitmap,
                        SpriteRectangle,
                        0,
                        0,
                        SpriteBitmap.Width, SpriteBitmap.Height,
                        GraphicsUnit.Pixel,
                        transparentAttributes);
                    state = AlienSpriteStates.visible;
                    break;
                case AlienSpriteStates.exploding:
                    ExplosionAnimatedBitmap.Draw(g, SpriteRectangle);
                    ExplosionAnimatedBitmap.FrameNo++;

                    if (ExplosionAnimatedBitmap.FrameNo == 0)
                    {
                        State = AlienSpriteStates.invisible;
                        IsAlive = false;
                    }
                    break;
                case AlienSpriteStates.invisible:
                    break;
            }
        }

        public override void Reset()
        {
            base.Reset();
            hitCount = 0;
            state = AlienSpriteStates.invisible;
        }


        public ImageAlienSprite(
            Bitmap inSpriteBitmap,
            Bitmap inExplodeBitmap,
            Sound inHitSound,
            Sound inExplodeSound,
			Rectangle inBoundingRectangle ) : base ( inExplodeBitmap, inHitSound,inExplodeSound,inBoundingRectangle)
        {
            SpriteBitmap = inSpriteBitmap;
            Reset();
        }
    }

    #endregion

}
