using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace SmartphoneSnake
{
    public partial class SnakeForm : Form
    {
        public SnakeForm()
        {
            init();
            InitializeComponent();
        }


        #region Snake

        #region Snake Position Variables

        /// <summary>
        /// Limit for the snake X value. Width of the screen
        /// minus the diameter of the snake dot.
        /// </summary>
        private float snakeXLimit;

        /// <summary>
        /// Limit for the snake Y value. Height of the screen 
        /// minus the diameter of the snake dot.
        /// </summary>
        private float snakeYLimit;

        /// <summary>
        /// Diameter of a dot in the snake. Set at a 
        /// DotWidthFactorth of the display width 
        /// - rounded down
        /// </summary>
        private int segmentDiameter;

        /// <summary>
        /// Size of the snake segments relative to the screen width.
        /// A value of n means that n segments would fit in the screen
        /// width. 
        /// </summary>
        private const int segmentWidthFactor = 25;

        /// <summary>
        /// Speed of snake in x direction.
        /// </summary>
        private float snakeXSpeed;

        /// <summary>
        /// Speed of snake in y direction.
        /// </summary>
        private float snakeYSpeed;

        /// <summary>
        /// The start speed of the snake.
        /// </summary>
        private float initialSnakeSpeed;

        /// <summary>
        /// The maximum speed of the snake.
        /// </summary>
        private float maximumSnakeSpeed;

        /// <summary>
        /// The change in speed that is applied to the snake.
        /// </summary>
        private float snakeSpeedStep;

        /// <summary>
        /// The speed of movement of the snake itself.
        /// </summary>
        private float snakeSpeed;

        #endregion 

        #region Snake Management

        #region Snake classes

        /// <summary>
        /// Keeps track of the position of a segment of the snake.
        /// </summary>
        class SnakeSegment
        {
            /// <summary>
            /// X position on the screen of the snake segment.
            /// </summary>
            public float X;

            /// <summary>
            /// Y position on the screen of the snake segment.
            /// </summary>
            public float Y;

            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float XSpeed;

            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float YSpeed;

            /// <summary>
            /// X offset to draw the snake segement (allows for segments
            /// of different sizes).
            /// </summary>
            public float XDrawOffset;

            /// <summary>
            /// Y offset to draw the snake segement (allows for segments
            /// of different sizes).
            /// </summary>
            public float YDrawOffset;

            /// <summary>
            /// Brush used to draw this snake segment.
            /// </summary>
            public Brush SnakeSegmentBrush;

            /// <summary>
            /// Rectangle defining the position of the snake on the screen.
            /// </summary>
            public Rectangle Rect;

            /// <summary>
            /// Draws the segment on the given graphics device
            /// </summary>
            /// <param name="g">device to be used for drawing</param>
            public void Draw ( Graphics g ) {
                Rect.X = (int)(X - XDrawOffset);
                Rect.Y = (int)(Y - YDrawOffset);
                g.FillRectangle(SnakeSegmentBrush, Rect);
            }

            /// <summary>
            /// Constructor. Do not set the position, since that
            /// will be set when the snake is placed on the screen.
            /// </summary>
            /// <param name="inBrush">brush to draw the snake segment with.</param>
            /// <param name="inDiameter">diameter of the snake segment.</param>
            public SnakeSegment(Brush inBrush, int inDiameter)
            {
                SnakeSegmentBrush = inBrush;
                XDrawOffset = inDiameter / 2.0f;
                YDrawOffset = inDiameter / 2.0f;
                Rect = new Rectangle(0, 0, inDiameter, inDiameter);
            }
        }

        class Snake
        {
            /// <summary>
            /// X coordinate of top left hand corner of the
            /// snake head. Origin at left.
            /// </summary>
            public float X;

            /// <summary>
            /// Y coordinate of top left hand corner of the 
            /// snake head. Origin at top.
            /// </summary>
            public float Y;

            /// <summary>
            /// The longest the snake is allowed to get in segments.
            /// </summary>
            private int maxSnakeLength ;

            /// <summary>
            /// The number of updates between each grow action.
            /// </summary>
            private int growSpeed;

            /// <summary>
            /// Counter for snake growing. As the time goes by the snake
            /// will grow. growCount is updated each time the snake is
            /// updated. When it reaches growSpeed the snake length
            /// is increased.
            /// </summary>
            private int growCount;

            /// <summary>
            /// The length of the snake at the start of the game.
            /// </summary>
            public int startSnakeLength ;

            /// <summary>
            /// The current length of the snake in segments.
            /// </summary>
            public int currentSnakeLength = 0;

            /// <summary>
            /// The entire snake. Set up when the program is
            /// initialised. Use increasing amounts of the snake
            /// as the game progresses.
            /// </summary>
            private SnakeSegment[] wholeSnake;

            /// <summary>
            /// Erase the snake from the screen. Draw each block with a 
            /// background colour.
            /// </summary>
            /// <param name="g">Graphics context to draw on.</param>
            public void Erase(Graphics g, Brush backgroundBrush)
            {
                for (int segNo = 0; segNo < currentSnakeLength; segNo++)
                {
                    SnakeSegment segment = wholeSnake[segNo];
                    g.FillRectangle(backgroundBrush, segment.Rect);
                }
            }

            /// <summary>
            /// Draw the snake on the graphics surface given.
            /// </summary>
            /// <param name="g">Graphics context to draw on.</param>
            public void Draw(Graphics g)
            {
                for (int segNo = 0; segNo < currentSnakeLength; segNo++)
                {
                    wholeSnake[segNo].Draw(g);
                }
            }

            /// <summary>
            /// Updates the snake.
            /// </summary>
            /// <param name="xSpeed">The X distance to move the snake head.</param>
            /// <param name="ySpeed">The Y distance to move the snake head/</param>
            /// <returns>True if the update succeeded. 
            /// False if the snake crashes into itself.</returns>
            public bool Update(float xSpeed, float ySpeed)
            {

                if (currentSnakeLength < startSnakeLength)
                {
                    wholeSnake[currentSnakeLength].XSpeed = 0.0f;
                    wholeSnake[currentSnakeLength].YSpeed = 0.0f;
                    currentSnakeLength++;
                }
                else
                {
                    if (++growCount >= growSpeed)
                    {
                        growCount = 0;
                        if (currentSnakeLength < maxSnakeLength)
                        {
                            wholeSnake[currentSnakeLength].XSpeed = 0.0f;
                            wholeSnake[currentSnakeLength].YSpeed = 0.0f;
                            currentSnakeLength++;
                        }
                    }
                }

                X += xSpeed;
                Y += ySpeed;

                SnakeSegment head = wholeSnake[0];

                for (int segNo = currentSnakeLength - 1; segNo > 0; segNo--)
                {
                    SnakeSegment segment = wholeSnake[segNo];
                    SnakeSegment segmentInFront = wholeSnake[segNo - 1];
                    if (head.Rect.IntersectsWith(segment.Rect))
                    {
                        ///only have a crash if the head segment is 
                        ///moving towards this segment
                        ///
                        if (((head.X > segment.X) && (xSpeed < 0)) ||
                             ((head.X < segment.X) && (xSpeed > 0)) ||
                             ((head.Y > segment.Y) && (ySpeed < 0)) ||
                             ((head.Y < segment.Y) && (ySpeed > 0)))
                        {
                            return false;
                        }
                    }
                    segmentInFront = wholeSnake[segNo - 1];
                    segment.X = segmentInFront.X;
                    segment.Y = segmentInFront.Y;
                    segment.XSpeed = segmentInFront.XSpeed;
                    segment.YSpeed = segmentInFront.YSpeed;
                }

                head.X = X;
                head.Y = Y;
                head.XSpeed = xSpeed;
                head.YSpeed = ySpeed;

                return true;
            }

            /// <summary>
            /// Used to see if any part of the snake intersects with any a given part of 
            /// the playfield. Used for placing bonus items or detecting hits.
            /// </summary>
            /// <param name="target">the rectangle to be tested</param>
            /// <returns>true if any part of the snake intersects with the rectangle</returns>
            public bool IntersectsWith(Rectangle target)
            {
                foreach (SnakeSegment segment in wholeSnake)
                {
                    if (target.IntersectsWith(segment.Rect)) {
                        return true;
                    }
                }
                return false;
            }

            /// <summary>
            /// Resets the snake for the next game. Places the snake on the
            /// screen at the given position and makes it zero length. It
            /// will grow to the start length as it is redrawn.
            /// </summary>
            /// <param name="inSnakeX">start X position, in pixels</param>
            /// <param name="inSnakeY">start Y position, in pixels</param>
            /// <param name="inStartSnakeLength">length to grow to at the start.</param>
            /// <param name="inMaxSnakeLength">the maximum length the snake is to have</param>
            /// <param name="inGrowSpeed">the number of updates between grows</param>
            public void ResetSnake(
                int inSnakeX, int inSnakeY,
                int inStartSnakeLength, 
                int inMaxSnakeLength,
                int inGrowSpeed)
            {
                X = inSnakeX;
                Y = inSnakeY;
                startSnakeLength = inStartSnakeLength;
                maxSnakeLength = inMaxSnakeLength;
                growSpeed = inGrowSpeed;
                growCount = 0;
                currentSnakeLength = 0;
            }

            public Snake(int inMaxSnakeLength, int dotDiameter, Brush snakeBrush)
            {
                maxSnakeLength = inMaxSnakeLength;

                wholeSnake = new SnakeSegment[maxSnakeLength];

                wholeSnake[0] = new SnakeSegment(snakeBrush, dotDiameter * 2);

                for (int segNo = 1; segNo < maxSnakeLength; segNo++)
                {
                    wholeSnake[segNo] = new SnakeSegment(snakeBrush, dotDiameter); ;
                }
            }
        }

        #endregion 

        #endregion

        #region Snake Drawing variables

        /// <summary>
        /// The snake in our game.
        /// </summary>
        private Snake snake;

        /// <summary>
        /// Brush used to draw the background. Kept as a member
        /// to avoid repeated recreation. Created in init.
        /// </summary>
        private Brush backgroundBrush;

        /// <summary>
        /// Graphics device for this form. Used to draw all
        /// screen items. Assigned in init.
        /// </summary>
        private Graphics gameGraphics;

        /// <summary>
        /// The background image for this game.
        /// </summary>
        private Image backgroundImage;

        /// <summary>
        /// now implementing double buffering
        /// </summary>
        private Image backBuffer;

        public Color TransparentColor = Color.FromArgb(254, 0, 254);

        public System.Drawing.Imaging.ImageAttributes TransparentAttributes;

        #endregion

        #endregion 

        #region Bonus

        #region Bonus Position Variables

        /// <summary>
        /// Size of the bonus relative to the screen width.
        /// A value of n means that n bonus items would fit in the screen
        /// width. 
        /// </summary>
        private int bonusWidthFactor = 10;

        /// <summary>
        /// Width of the bonus, in pixels. Calculated in the init method.
        /// </summary>
        private int bonusWidth;

        #endregion

        #region Bonus class

        /// <summary>
        /// Holds information about a bonus item on the screen
        /// </summary>
        class Bonus
        {
            /// <summary>
            /// True if the bonus is presently enabled
            /// </summary>
            public bool Enabled;

            /// <summary>
            /// Position and size on the screen
            /// </summary>
            public Rectangle Rect;

            /// <summary>
            /// Brush used to draw the bonus item
            /// </summary>
            public Brush BonusBrush;

            /// <summary>
            /// Value gained when the bonus is eaten.
            /// </summary>
            public int Value;

            /// <summary>
            /// Random number generator to use to place food.
            /// </summary>
            public Random Rand;

            /// <summary>
            /// Create a new bonus item.
            /// </summary>
            /// <param name="inDiameter">Size of the bonus.</param>
            /// <param name="inBonusBrush">Brush to draw it with.</param>
            /// <param name="inValue">The amount the bonus is worth</param>
            /// <param name="inRand">The random number generator to use.</param>
            public Bonus(int inDiameter, Brush inBonusBrush, int inValue, Random inRand)
            {
                Rect = new Rectangle(0, 0, inDiameter, inDiameter);
                BonusBrush = inBonusBrush;
                Value = inValue;
                Rand = inRand;
            }

            /// <summary>
            /// Places the bonus on the screen in a place which does not
            /// intersect with the given snake.
            /// </summary>
            /// <param name="percentLikely">The percentage likeliness of gettting a 
            /// bonus. The bigger the number, up to 100, the likelier the 
            /// bonus.</param>
            /// <param name="snake">The snake we must not intersect with</param>
            /// <returns>true if the placing worked. False if not.</returns>
            /// <param name="maxX">Maximum X value for the position</param>
            /// <param name="maxY">Maximum Y value for the position</param>
            public bool PlaceBonus(int percentLikely, Snake snake, int maxX, int maxY)
            {
                /// Don't place the bonus if it is already there
                if (Enabled)
                {
                    return true;
                }

                /// Only appear every now and then
                if (Rand.Next(100) > percentLikely)
                {
                    return false;
                }

                /// Try 5 times.
                for (int i = 0; i < 5; i++)
                {
                    Rect.X = Rand.Next(maxX - Rect.Width);
                    Rect.Y = Rand.Next(maxY - Rect.Height);
                    if ( !snake.IntersectsWith(Rect) ) {
                        Enabled = true;
                        return true;
                    }
                }
                Enabled = false;
                return false;
            }

            /// <summary>
            /// Check to see if the bonus and the snake intersect.
            /// </summary>
            /// <param name="snake">snake to check against</param>
            /// <returns>true if the bonus has been won</returns>
            public bool CheckBonus(Snake snake)
            {
                if( snake.IntersectsWith (Rect ) ) {
                    Enabled = false;
                    return true;
                }
                return false;
            }

            /// <summary>
            /// Draw the bonus on the specified graphics surface.
            /// </summary>
            /// <param name="g"></param>
            public void Draw ( Graphics g ) {
                if (Enabled)
                {
                    g.FillRectangle(BonusBrush, Rect);
                }
            }


        }

        #endregion

        #region Bonus Management

        /// <summary>
        /// The one and only bonus value.
        /// </summary>
        Bonus bonus;

        #endregion 

        #endregion

        #region Game State variables

        /// <summary>
        /// When true the game is being played.
        /// </summary>
        private bool gameActive = false;

        /// <summary>
        /// When true the game is in pause state.
        /// </summary>
        private bool gamePaused = false;

        #endregion

        /// <summary>
        /// Sets up the game at the start.
        /// </summary>
        private void init()
        {
            backgroundBrush = new SolidBrush(Color.White);

            backBuffer = new Bitmap(this.ClientRectangle.Width, this.ClientRectangle.Height);
            backgroundImage = new Bitmap(backBuffer);

            // Get our assembly, so that we can get the grass out of it
            System.Reflection.Assembly execAssem =
                System.Reflection.Assembly.GetExecutingAssembly();

            // fetch the grass
            Bitmap grassImage = new System.Drawing.Bitmap(execAssem.GetManifestResourceStream(@"SmartphoneSnake.Images.grass.jpg"));

            // now scale the grass image and draw onto our background

            using (Graphics backGraph = Graphics.FromImage(backgroundImage))
            {
                Rectangle source = new Rectangle (0, 0, grassImage.Width, grassImage.Height);
                Rectangle dest = new Rectangle (0,0, backBuffer.Width, backBuffer.Height);
                backGraph.DrawImage(grassImage, dest, source, GraphicsUnit.Pixel);
            }

            gameGraphics = Graphics.FromImage(backBuffer);

            segmentDiameter = this.ClientRectangle.Width / segmentWidthFactor;

            snake = new Snake(200, segmentDiameter, new SolidBrush(Color.Red));

            snake.ResetSnake(
                (this.ClientRectangle.Width - segmentDiameter) / 2, // X position of head
                (this.ClientRectangle.Height - segmentDiameter) / 2, // Y position of head
                10,
                200,
                5);

            initialSnakeSpeed = (float)segmentDiameter / 1.1f;
            maximumSnakeSpeed = (float)segmentDiameter / 2.0f;
            snakeSpeedStep = (maximumSnakeSpeed - initialSnakeSpeed) / 5;

            snakeSpeed = initialSnakeSpeed;

            snakeXSpeed = snakeSpeed;
            snakeYSpeed = 0;

            snakeXLimit = this.ClientRectangle.Width - segmentDiameter;
            snakeYLimit = this.ClientRectangle.Height - segmentDiameter;

            bonusWidth = this.ClientRectangle.Width / bonusWidthFactor;

            bonus = new Bonus(bonusWidth, new SolidBrush(Color.Yellow), 50, new System.Random());
        }

        /// <summary>
        /// draw the game screen using the graphics for this form.
        /// </summary>
        private void drawGameScreen()
        {
            snake.Draw(gameGraphics);
            bonus.Draw(gameGraphics);
        }

        /// <summary>
        /// Erases the game screen using the graphics for this form.
        /// </summary>
        private void eraseGameScreen()
        {
            gameGraphics.DrawImage(backgroundImage, 0, 0);
        }

        #region Timer routines

        /// <summary>
        /// The tick count recorded at the start of the timing period.
        /// </summary>
        private int clockStartTick;

        /// <summary>
        /// Counter for the number of readings to be taken.
        /// </summary>
        private int readingCount = 0;

        /// <summary>
        /// Number of readings to be averaged over.
        /// </summary>
        private int sampleSize = 20;

        /// <summary>
        /// Total of readings taken.
        /// </summary>
        private float readingTotal = 0.0f;

        /// <summary>
        /// Calculated average result.
        /// </summary>
        private float timerResult;

        /// <summary>
        /// Starts the timer ticking.
        /// </summary>
        private void startTimer()
        {
            clockStartTick = System.Environment.TickCount;
        }

        /// <summary>
        /// Obtains a new time value, adds it to the total and
        /// displays the average if we are at the end of a sample
        /// period.
        /// </summary>
        private void showTimer()
        {
            readingTotal += (System.Environment.TickCount - clockStartTick);

            if (++readingCount == sampleSize)
            {
                timerResult = readingTotal / sampleSize;
                this.Text = "Time : " + timerResult.ToString();
                readingTotal = 0.0f;
                readingCount = 0;
            }
        }

        #endregion

        #region Game Play


        /// <summary>
        /// Update the snake position. End the game if
        /// it hits anything.
        /// </summary>
        private void updateGame()
        {
            if ( !snake.Update(snakeXSpeed, snakeYSpeed) ||
                 (snake.X < 0.0f) || (snake.X > snakeXLimit) ||
                 (snake.Y < 0.0f) || (snake.Y > snakeYLimit)
               )
            {
                MessageBox.Show("Game Over");
                eraseGameScreen();
                gameActive = false;
            }
            if (bonus.Enabled)
            {
                bonus.CheckBonus(snake);
            }
            else
            {
                bonus.PlaceBonus(20, snake, this.Width, this.Height);
            }
        }


        /// <summary>
        /// Plays the game. Sets gameActive to true and
        /// updates the game until it becomes false.
        /// </summary>
        private void playGame()
        {
            init();
            gameActive = true;
            while (gameActive)
            {
                startTimer();
                if (!gamePaused)
                {
                    eraseGameScreen();
                    updateGame();
                    drawGameScreen();
                }
                showTimer();
                System.Threading.Thread.Sleep(100);
                Application.DoEvents();
                Invalidate();
            }
        }

        /// <summary>
        /// Refers to a method with the signature required for 
        /// Form.Invoke calls.
        /// </summary>
        private delegate void gameMethod();

        /// <summary>
        /// Refers to the method which will play the game. Set
        /// to playGame when the game starts.
        /// </summary>
        gameMethod playDelegate;

        /// <summary>
        /// Starts a new game.
        /// </summary>
        private void startGame()
        {
            playDelegate = new gameMethod(playGame);
            Invoke(playDelegate);
        }

        /// <summary>
        /// Pause the game.
        /// </summary>
        private void pauseGame()
        {
            gamePaused = true;
        }

        /// <summary>
        /// Resume the game
        /// </summary>
        private void resumeGame()
        {
            gamePaused = false;
        }

        #endregion 

        #region Event Handlers

        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);
            pauseGame();
        }

        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);
            resumeGame();
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            e.Graphics.DrawImage(backBuffer, 0, 0);
        }

        private void playMenuItem_Click(object sender, EventArgs e)
        {
            startGame();
        }

        private void SnakeForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((e.KeyCode == System.Windows.Forms.Keys.F1))
            {
                // Soft Key 1
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F2))
            {
                // Soft Key 2
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Up))
            {
                // Up
                snakeXSpeed = 0.0f;
                snakeYSpeed = -snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Down))
            {
                // Down
                snakeXSpeed = 0.0f;
                snakeYSpeed = snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Left))
            {
                // Left
                snakeXSpeed = -snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Right))
            {
                // Right
                snakeXSpeed = snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Enter))
            {
                // Enter
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D1))
            {
                // 1
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D2))
            {
                // 2
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D3))
            {
                // 3
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D4))
            {
                // 4
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D5))
            {
                // 5
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D6))
            {
                // 6
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D7))
            {
                // 7
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D8))
            {
                // 8
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D9))
            {
                // 9
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F8))
            {
                // *
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D0))
            {
                // 0
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F9))
            {
                // #
            }
        }

        #endregion 
    }
}